mod imp;

use adw::{MessageDialog, NavigationDirection, ResponseAppearance};
use crate::ui::objects::app::{AppData, AppObject};

use adw::prelude::*;
use adw::subclass::prelude::*;
use f_uuid::Uuid;
use gtk::{gio, glib, Entry, ListBoxRow, Label, pango, SelectionMode, Button, Orientation, CenterBox};
use glib::{clone, Object};

glib::wrapper! {
    pub struct Window(ObjectSubclass<imp::Window>)
        @extends adw::ApplicationWindow, gtk::Window, gtk::Widget,
        @implements gio::ActionGroup, gio::ActionMap, gtk::Accessible, gtk::Buildable,
                    gtk::ConstraintTarget, gtk::Native, gtk::Root, gtk::ShortcutManager;
}

impl Window {
    pub fn new(app: &adw::Application) -> Self {
        Object::builder().property("application", app).build()
    }

    fn apps(&self) -> gio::ListStore {
        self.imp()
            .apps
            .get()
            .expect("`apps` should be set in `set_apps`.")
            .clone()
    }

    fn create_app_row(&self, app_object: &AppObject) -> ListBoxRow {
        let label = Label::builder()
            .ellipsize(pango::EllipsizeMode::End)
            .xalign(0.0)
            .build();

        let remove_button = Button::builder()
            .icon_name("user-trash-symbolic")
            .tooltip_text("Remove the application")
            .action_name("win.remove-application")
            .build();

        let bbox = CenterBox::builder()
            .orientation(Orientation::Horizontal)
            .build();

        bbox.set_start_widget(Some(&label));
        bbox.set_end_widget(Some(&remove_button));

        self.imp().apps_list.remove(&bbox);

        app_object
            .bind_property("name", &label, "label")
            .sync_create()
            .build();

        ListBoxRow::builder().child(&bbox).build()
    }

    fn current_app(&self) -> AppObject {
        self.imp()
            .current_app
            .borrow()
            .clone()
            .expect("`current_app` should be set in `set_current_app`.")
    }

    fn new_application(&self) {
        let entry = Entry::builder()
            .placeholder_text("Name")
            .activates_default(true)
            .build();

        let cancel_response = "cancel";
        let create_response = "create";

        let dialog = MessageDialog::builder()
            .heading("New Application")
            .transient_for(self)
            .modal(true)
            .destroy_with_parent(true)
            .close_response(cancel_response)
            .default_response(create_response)
            .extra_child(&entry)
            .build();

        dialog.add_responses(&[(cancel_response, "Cancel"), (create_response, "Create")]);
        dialog.set_response_enabled(create_response, false);
        dialog.set_response_appearance(create_response, ResponseAppearance::Suggested);

        entry.connect_changed(clone!(@weak dialog => move |entry| {
            let text = entry.text();
            let empty = text.is_empty();

            dialog.set_response_enabled(create_response, !empty);

            if empty {
                entry.add_css_class("error");
            } else {
                entry.remove_css_class("error");
            }
        }));

        dialog.connect_response(
            None,
            clone!(@weak self as window, @weak entry => move |dialog, response| {
                dialog.destroy();

                if response != create_response {
                    return;
                }

                let name = entry.text().to_string();
                let app = AppObject::new(&name, "/some/path");

                window.apps().append(&app);
                window.set_current_app(app);

                window.imp().leaflet.navigate(NavigationDirection::Forward);
            }),
        );

        dialog.present();
    }

    fn remove_application(&self, id: &str) {
        let apps = self.imp().apps.get().unwrap();
        let index = apps.find_with_equal_func(|obj| -> bool {
            let id_prop = obj
                .downcast_ref::<AppObject>()
                .expect("The object should be of type 'AppObject'")
                .property_value("id");

            let app_id = id_prop.get::<String>().expect("AppObject 'id' property should be a string");
            app_id == id
        });

        dbg!(index);
    }

    fn restore_data(&self) {
        let data = vec!(
            AppData {
                id: Uuid::new_random(),
                name: "IntelliJ".to_string(),
                path: "/home/william/.config/jetbrains/intellij".to_string(),
            },
            AppData {
                id: Uuid::new_random(),
                name: "Spotify".to_string(),
                path: "/home/william/.config/spotify".to_string(),
            });

        let apps: Vec<AppObject> = data.iter()
            .map(AppObject::from_app_data)
            .collect();

        self.apps().extend_from_slice(&apps);

        if let Some(app) = apps.first() {
            self.set_current_app(app.clone());
        }
    }

    fn select_app_row(&self) {
        if let Some(index) = self.apps().find(&self.current_app()) {
            let row = self.imp().apps_list.row_at_index(index as i32);
            self.imp().apps_list.select_row(row.as_ref());
        }
    }

    fn set_current_app(&self, app: AppObject) {
        self.imp().app_view.set_application(app.clone());
        self.imp().current_app.replace(Some(app));
        self.select_app_row();
    }

    fn set_stack(&self) {
        if self.apps().n_items() > 0 {
            self.imp().stack.set_visible_child_name("main");
        } else {
            self.imp().stack.set_visible_child_name("empty");
        }
    }

    fn setup_actions(&self) {
        let action_new_application = gio::SimpleAction::new("new-application", None);
        action_new_application.connect_activate(clone!(@weak self as window => move |_, _| {
            window.new_application();
        }));
        self.add_action(&action_new_application);
    }

    fn setup_apps(&self) {
        let apps = gio::ListStore::new::<AppObject>();
        self.imp()
            .apps
            .set(apps.clone())
            .expect("Could not set apps");

        self.imp().apps_list.bind_model(
            Some(&apps),
            clone!(@weak self as window => @default-panic, move |obj| {
                let app_object = obj
                    .downcast_ref()
                    .expect("The object should be of type `AppObject`.");
                let row = window.create_app_row(app_object);
                row.upcast()
            }),
        )
    }

    fn setup_callbacks(&self) {
        self.set_stack();
        self.apps().connect_items_changed(
            clone!(@weak self as window => move |_, _, _, _| {
                window.set_stack();
            })
        );

        self.imp().apps_list.connect_row_activated(
            clone!(@weak self as window => move |_, row| {
                let index = row.index();
                let selected_app = window.apps()
                    .item(index as u32)
                    .expect("There needs to be an object at this position.")
                    .downcast::<AppObject>()
                    .expect("The object needs to be a 'AppObject'.");
                window.set_current_app(selected_app);
                window.imp().leaflet.navigate(NavigationDirection::Forward);
            })
        );

        self.imp().leaflet.connect_folded_notify(
            clone!(@weak self as window => move |leaflet| {
                if leaflet.is_folded() {
                    window
                        .imp()
                        .apps_list
                        .set_selection_mode(SelectionMode::None)
                } else {
                    window
                        .imp()
                        .apps_list
                        .set_selection_mode(SelectionMode::Single);
                    window.select_app_row();
                }
            })
        );
    }
}
